@@ -4,6 +4,7 @@ use std::fmt::Write;
 use std::path::Path;
 
 const INDENT: &str = "  ";
+const DEFAULT_COLUMNS: usize = 16;
 
 fn basename(path: &str) -> &str {
     // Wish expect() could use String. This is messy.
@@ -42,6 +43,18 @@ impl Indentable for str {
     }
 }
 
+macro_rules! hex_dualfmt {
+    ($id:expr) => {
+        format!("<span class='number' title='{}'>{:#x}</span>", $id, $id)
+    };
+}
+
+macro_rules! dec_dualfmt {
+    ($id:expr) => {
+        format!("<span class='number' title='{:#x}'>{}</span>", $id, $id)
+    };
+}
+
 macro_rules! wrow {
     ($dst:expr, $indent_level:expr, $lhs:expr, $rhs:expr) => {
         wnonl!($dst, $indent_level, "<tr> ");
@@ -67,10 +80,11 @@ macro_rules! wnonl {
 
 fn generate_head(o: &mut String, elf: &ParsedElf) {
     let stylesheet: String = include_str!("style.css").indent_lines(3);
+    let viewport = "width=900, initial-scale=1";
 
     w!(o, 1, "<head>");
     w!(o, 2, "<meta charset='utf-8'>");
-    w!(o, 2, "<meta name='viewport' content='width=900, initial-scale=1'>");
+    w!(o, 2, "<meta name='viewport' content='{}'>", viewport);
     w!(o, 2, "<title>{}</title>", basename(&elf.filename));
     w!(o, 2, "<style>");
     wnonl!(o, 0, "{}", stylesheet);
@@ -101,30 +115,76 @@ fn generate_svg_element(o: &mut String) {
 }
 
 fn generate_file_info_table(o: &mut String, elf: &ParsedElf) {
-    w!(o, 4, "<table>");
+    w!(o, 2, "<table>");
 
     for (id, desc, value) in elf.information.iter() {
-        wnonl!(o, 5, "<tr id='info_{}'> ", id);
+        wnonl!(o, 3, "<tr id='info_{}'> ", id);
         wnonl!(o, 0, "<td>{}:</td> ", desc);
         wnonl!(o, 0, "<td>{}</td> ", value);
         w!(o, 0, "</tr>");
     }
 
-    w!(o, 4, "</table>");
+    w!(o, 2, "</table>");
+}
+
+fn generate_help(o: &mut String) {
+    let legend_items = [
+        ("ident", "ELF Identification"),
+        ("ehdr", "ELF Header"),
+        ("phdr", "Program Header"),
+        ("shdr", "Section Header"),
+        ("segment", "Segment"),
+        ("section", "Section"),
+        ("segm_sect_legend", "Segment &amp; Section overlap"),
+    ];
+    let help_text = "The leftmost column shows offsets within the file. \
+                     The middle column is the file dump. It has ELF structs, sections and segments \
+                     highlighted. Some fields that reference areas in the file are clickable and \
+                     connected with arrows. The rightmost column shows printable ASCII characters \
+                     corresponding to the file bytes.";
+
+    w!(o, 4, "<p>{}</p>", help_text);
+
+    w!(o, 4, "<p>Legend</p>");
+    w!(o, 4, "<ul>");
+    for (class, desc) in legend_items {
+        w!(o, 5, "<li><span class='legend_rect {}'></span>{}</li>", class, desc);
+    }
+    w!(o, 4, "</ul>");
+}
+
+fn generate_right_menu(o: &mut String) {
+    let credits = format!("generated with elfcat {}", env!("CARGO_PKG_VERSION"));
+    let url = "https://github.com/ruslashev/elfcat";
+
+    w!(o, 3, "<a id='credits' href='{}'>{}</a>", url, credits);
+
+    w!(o, 3, "<button class='textbutton' id='settings_toggle'>Settings</button>");
+
+    w!(o, 3, "<button class='textbutton' id='help_toggle'>Help</button>");
+
+    w!(o, 3, "<div class='right_hidden' id='settings'>");
+    w!(o, 4, "<label for='arrow_opacity_range'>Arrow opacity:</label>");
+    w!(o, 4, "<input type='range' id='arrow_opacity_range' min='0' max='100' value='100'>");
+    w!(o, 3, "</div>");
+
+    w!(o, 3, "<div class='right_hidden' id='help'>");
+    generate_help(o);
+    w!(o, 3, "</div>");
 }
 
 fn generate_phdr_info_table(o: &mut String, phdr: &ParsedPhdr, idx: usize) {
     let items = [
         ("Type", &ptype_to_string(phdr.ptype)),
         ("Flags", &phdr.flags),
-        ("Offset in file", &format!("{}", phdr.file_offset)),
-        ("Size in file", &format!("{}", phdr.file_size)),
-        ("Vaddr in memory", &format!("{:#x}", phdr.vaddr)),
-        ("Size in memory", &format!("{:#x}", phdr.memsz)),
-        ("Alignment", &format!("{:#x}", phdr.alignment)),
+        ("Offset in file", &hex_dualfmt!(phdr.file_offset)),
+        ("Size in file", &dec_dualfmt!(phdr.file_size)),
+        ("Vaddr in memory", &hex_dualfmt!(phdr.vaddr)),
+        ("Size in memory", &hex_dualfmt!(phdr.memsz)),
+        ("Alignment", &hex_dualfmt!(phdr.alignment)),
     ];
 
-    w!(o, 5, "<table class='conceal' id='info_phdr{}'>", idx);
+    w!(o, 5, "<table class='conceal valign_top' id='info_phdr{}'>", idx);
 
     for (desc, value) in items.iter() {
         wrow!(o, 6, desc, value);
@@ -144,16 +204,16 @@ fn generate_shdr_info_table(o: &mut String, elf: &ParsedElf, shdr: &ParsedShdr,
         ("Name", elf.shnstrtab.get(shdr.name)),
         ("Type", &shtype_to_string(shdr.shtype)),
         ("Flags", &shflags_to_string(shdr.flags)),
-        ("Vaddr in memory", &format!("{:#x}", shdr.addr)),
-        ("Offset in file", &format!("{}", shdr.file_offset)),
-        ("Size in file", &format!("{}", shdr.size)),
+        ("Vaddr in memory", &hex_dualfmt!(shdr.addr)),
+        ("Offset in file", &hex_dualfmt!(shdr.file_offset)),
+        ("Size in file", &dec_dualfmt!(shdr.size)),
         ("Linked section", &format!("{}", shdr.link)),
-        ("Extra info", &format!("{}", shdr.link)),
-        ("Alignment", &format!("{:#x}", shdr.addralign)),
-        ("Size of entries", &format!("{}", shdr.entsize)),
+        ("Extra info", &dec_dualfmt!(shdr.link)),
+        ("Alignment", &hex_dualfmt!(shdr.addralign)),
+        ("Size of entries", &dec_dualfmt!(shdr.entsize)),
     ];
 
-    w!(o, 5, "<table class='conceal' id='info_shdr{}'>", idx);
+    w!(o, 5, "<table class='conceal valign_top' id='info_shdr{}'>", idx);
 
     for (desc, value) in items.iter() {
         wrow!(o, 6, desc, value);
@@ -294,10 +354,10 @@ fn has_section_detail(ptype: u32) -> bool {
 
 fn generate_segment_info_tables(o: &mut String, elf: &ParsedElf) {
     for (idx, phdr) in elf.phdrs.iter().enumerate() {
-        w!(o, 5, "<table class='conceal' id='info_segment{}'>", idx);
+        w!(o, 5, "<table class='conceal valign_top' id='info_segment{}'>", idx);
         wrow!(o, 6, "Segment type", &ptype_to_string(phdr.ptype));
-        wrow!(o, 6, "Size in file", phdr.file_size);
-        wrow!(o, 6, "Size in memory", phdr.memsz);
+        wrow!(o, 6, "Size in file", dec_dualfmt!(phdr.file_size));
+        wrow!(o, 6, "Size in memory", dec_dualfmt!(phdr.memsz));
 
         if has_segment_detail(phdr.ptype) {
             w!(o, 6, "<tr><td><br></td></tr>");
@@ -310,9 +370,9 @@ fn generate_segment_info_tables(o: &mut String, elf: &ParsedElf) {
 
 fn generate_section_info_tables(o: &mut String, elf: &ParsedElf) {
     for (idx, shdr) in elf.shdrs.iter().enumerate() {
-        w!(o, 5, "<table class='conceal' id='info_section{}'>", idx);
+        w!(o, 5, "<table class='conceal valign_top' id='info_section{}'>", idx);
         wrow!(o, 6, "Section type", &shtype_to_string(shdr.shtype));
-        wrow!(o, 6, "Size", shdr.size);
+        wrow!(o, 6, "Size", dec_dualfmt!(shdr.size));
 
         if has_section_detail(shdr.shtype) {
             w!(o, 6, "<tr><td><br></td></tr>");
@@ -379,12 +439,7 @@ fn add_highlight_script(o: &mut String) {
 fn add_description_script(o: &mut String) {
     w!(o, 2, "<script type='text/javascript'>");
 
-    wnonl!(
-        o,
-        0,
-        "{}",
-        include_str!("js/description.js").indent_lines(3)
-    );
+    wnonl!(o, 0, "{}", include_str!("js/description.js").indent_lines(3));
 
     w!(o, 2, "</script>");
 }
@@ -404,7 +459,7 @@ fn add_offsets_script(o: &mut String, elf: &ParsedElf) {
 
     wnonl!(o, 0, "{}", include_str!("js/offsets.js").indent_lines(3));
 
-    w!(o, 3, "populateOffsets(16)");
+    w!(o, 3, "populateOffsets({})", DEFAULT_COLUMNS);
 
     w!(o, 2, "</script>");
 }
@@ -418,23 +473,11 @@ fn add_arrows_script(o: &mut String, elf: &ParsedElf) {
     w!(o, 3, "connect('#e_shoff', '#bin_shdr0');");
 
     for i in 0..elf.phdrs.len() {
-        w!(
-            o,
-            3,
-            "connect('#bin_phdr{} > .p_offset', '#bin_segment{}');",
-            i,
-            i
-        );
+        w!(o, 3, "connect('#bin_phdr{} > .p_offset', '#bin_segment{}');", i, i);
     }
 
     for i in 0..elf.shdrs.len() {
-        w!(
-            o,
-            3,
-            "connect('#bin_shdr{} > .sh_offset', '#bin_section{}');",
-            i,
-            i
-        );
+        w!(o, 3, "connect('#bin_shdr{} > .sh_offset', '#bin_section{}');", i, i);
     }
 
     w!(o, 2, "</script>");
@@ -449,6 +492,14 @@ fn add_collapsible_script(o: &mut String) {
     w!(o, 2, "</script>");
 }
 
+fn add_settings_script(o: &mut String) {
+    w!(o, 2, "<script type='text/javascript'>");
+
+    wnonl!(o, 0, "{}", include_str!("js/settings.js").indent_lines(3));
+
+    w!(o, 2, "</script>");
+}
+
 fn add_scripts(o: &mut String, elf: &ParsedElf) {
     add_highlight_script(o);
 
@@ -462,6 +513,8 @@ fn add_scripts(o: &mut String, elf: &ParsedElf) {
     add_offsets_script(o, elf);
 
     add_arrows_script(o, elf);
+
+    add_settings_script(o);
 }
 
 fn format_magic(byte: u8) -> String {
@@ -511,7 +564,7 @@ fn generate_dump_for_byte(idx: usize, dump: &mut String, elf: &ParsedElf) {
         dump.push_str("</span>");
     }
 
-    dump.push_str(if (idx + 1) % 16 == 0 { "\n" } else { " " });
+    dump.push_str(if (idx + 1) % DEFAULT_COLUMNS == 0 { "\n" } else { " " });
 }
 
 // assumes balance == 1
@@ -565,13 +618,10 @@ fn generate_file_dump(elf: &ParsedElf) -> String {
         // disable while working on offsets
         if false && balance == 1 {
             if let Some(new_idx) = skip_bytes(i, len, elf) {
-                dump += format!(
-                    "<span {}>..</span>",
-                    elf.ranges.data[i][0].span_attributes()
-                )
-                .as_str();
+                dump +=
+                    format!("<span {}>..</span>", elf.ranges.data[i][0].span_attributes()).as_str();
 
-                dump += if (i + 1) % 16 == 0 { "\n" } else { " " };
+                dump += if (i + 1) % DEFAULT_COLUMNS == 0 { "\n" } else { " " };
 
                 i = new_idx;
                 continue;
@@ -602,7 +652,7 @@ fn generate_ascii_dump(o: &mut String, elf: &ParsedElf) {
             wnonl!(o, 0, ".");
         }
 
-        if (i + 1) % 16 == 0 {
+        if (i + 1) % DEFAULT_COLUMNS == 0 {
             w!(o, 0, "");
         }
     }
@@ -613,19 +663,11 @@ fn generate_body(o: &mut String, elf: &ParsedElf) {
 
     generate_svg_element(o);
 
-    w!(o, 2, "<table id='headertable'>");
-    w!(o, 3, "<td>");
+    w!(o, 2, "<div id='rightmenu'>");
+    generate_right_menu(o);
+    w!(o, 2, "</div>");
+
     generate_file_info_table(o, elf);
-    w!(o, 3, "</td>");
-    w!(o, 3, "<td id='rightmenu'>");
-    w!(
-        o,
-        4,
-        "<p id='credits'>generated with elfcat {}</p>",
-        env!("CARGO_PKG_VERSION")
-    );
-    w!(o, 3, "</td>");
-    w!(o, 2, "</table>");
 
     w!(o, 2, "<div id='offsets'></div>");
 
